@@ -24,13 +24,16 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * $Id: enable_pps.c,v 1.13 1999/01/14 21:11:40 windl Exp $ */
+ * $Id: enable_pps.c,v 1.13.2.1 1999/03/11 22:04:20 windl Exp $
+ */
 #include	<unistd.h>
 #include	<time.h>
 #include	<stdlib.h>
 #include	<stdio.h>
 #include	<string.h>
 #include	<errno.h>
+
+#define	NANO	/* prefer the Nanokernel if available */
 #include	<sys/timex.h>
 
 #include	<sys/ppsclock.h>
@@ -48,14 +51,42 @@
 				 */
 #endif
 
+#ifdef	STA_NANO		/* check if compiling with Nanokernel */
+#define	tv_frac	tv_nsec		/* using struct timespec */
+#else
+#define	tv_frac	tv_usec		/* using struct timeval */
+#endif
+
+#ifndef	SHIFT_USEC
+#define SHIFT_USEC	16	/* compatibility define */
+#define	PPS_SHIFT	4	/* compatibility define */
+#define	PPS_SHIFTMAX	8	/* compatibility define */
+#endif
+
+#ifndef	ADJ_ADJTIME
+#define ADJ_ADJTIME	ADJ_OFFSET_SINGLESHOT	/* compatibility define */
+#endif
+
 #define OFFSET_TOLERANCE	50000
 #define STABIL_TOLERANCE	(10 << SHIFT_USEC)
 #define	PPS_INSANE		(STA_PPSERROR|STA_PPSWANDER|STA_PPSJITTER)
 
 #define	ALL_BITS_ON(v, b)		(((v) & (b)) == (b))
 #define	ALL_BITS_OFF(v, b)		(((v) & (b)) == 0)
+#define	BITS_OFF(v, b)			((v) & ~(b))
+#define	BITS_ON(v, b)			((v) | (b))
 
-static	const char	id[] = "$Id: enable_pps.c,v 1.13 1999/01/14 21:11:40 windl Exp $";
+static	const char	id[] = "$Id: enable_pps.c,v 1.13.2.1 1999/03/11 22:04:20 windl Exp $";
+
+#define	BIT(n)	(1UL << (n))
+#define OPT_MONITOR		BIT(0)	/* perform monitor loop */
+#define OPT_RESET		BIT(1)	/* reset clock state */
+#define OPT_NO_OFFSET		BIT(2)	/* don't adjust offset */
+#define OPT_NO_STATUS		BIT(3)	/* don't adjust status */
+#define OPT_NO_TIMECONST	BIT(4)	/* don't adjust time constant */
+#define OPT_JITTER_LOOP		BIT(5)	/* monitor jitter only */
+
+static	unsigned long	options = OPT_MONITOR | OPT_RESET;
 
 static	char	polarity	= 'p';	/* p[ositive] | n[egative] | 0) */
 
@@ -74,6 +105,7 @@ static	void	print_header(void)
 	       "errcnt", "stbcnt");
 }
 
+/* (partially) print current time state */
 static	void	print_stats(const struct timex *tp)
 {
 	printf("%04x %7ld %7.3f %7.3f %4ld %1d %7.3f %4ld %4ld %4ld %4ld\n",
@@ -106,9 +138,9 @@ static	char const	*status_str[] = {	/* from <timex.h> */
 	"PPSERROR",	/* bit 11 */
 
 	"CLOCKERR",	/* bit 12 */
-	"???13",	/* bit 13 */
-	"???14",	/* bit 14 */
-	"???15",	/* bit 15 */
+	"NANO",		/* bit 13 */
+	"MODE",		/* bit 14 */
+	"CLK",		/* bit 15 */
 };
 
 /* write strings for set status bits on ``fp'' */
@@ -205,6 +237,8 @@ static	int	update_status(struct timex *txp, struct ppsclockev *evp,
 	int	change;
 	char	*sep = " (";
 
+	if ( options & OPT_NO_STATUS )
+		return(0);
 	txp->status |= set_bits;
 	txp->status &= ~clear_bits;
 	txp->modes = MOD_STATUS;
@@ -231,6 +265,8 @@ static	int	update_status(struct timex *txp, struct ppsclockev *evp,
 /* reset kernel PLL */
 static	void	restart(struct timex *txp, struct ppsclockev *evp)
 {
+	if ( ALL_BITS_OFF(options, OPT_RESET) )
+		return;
 	txp->modes = MOD_ESTERROR | MOD_MAXERROR;
 	/* indicate that we are trying to synchronize */
 	txp->esterror = txp->maxerror = MAXPHASE / 1000;
@@ -248,6 +284,8 @@ static	void	restart(struct timex *txp, struct ppsclockev *evp)
 /* use current time constant from PPS discipline if it is working properly */
 static	void	update_constant(struct timex *txp, struct ppsclockev *evp)
 {
+	if ( options & OPT_NO_TIMECONST )
+		return;
 	if ( (txp->status & STA_PPSSIGNAL) != 0 &&
 	     (txp->status & PPS_INSANE) == 0 )
 		txp->constant = txp->shift;		/* copy PPS PLL constant */
@@ -276,7 +314,7 @@ static	void	pll_adj_offset(struct timex *txp, struct ppsclockev *evp)
 	txp->esterror = txp->maxerror = offset >= 0 ? : -offset;
 	txp->offset = (txp->offset + offset) / 2;
 	if ( txp->offset > MAXPHASE / 2 || txp->offset < -MAXPHASE / 2 )
-		txp->modes = ADJ_OFFSET_SINGLESHOT;
+		txp->modes = ADJ_ADJTIME;
 	else
 		txp->modes |= MOD_OFFSET;
 	tdelta = t - last_adj_time;
@@ -297,6 +335,8 @@ static	void	pll_adj_offset(struct timex *txp, struct ppsclockev *evp)
 /* adjust offset if conditions are sane */
 static	void	update_offset(struct timex *txp, struct ppsclockev *evp)
 {
+	if ( options & OPT_NO_OFFSET )
+		return;
 	if ( ((txp->status & STA_PPSSIGNAL) != 0 &&
 	      (txp->status & PPS_INSANE) == 0) ||
 	     (txp->status & STA_PPSSIGNAL) == 0 )
@@ -309,14 +349,132 @@ static	void	status(char const *str)
 	fprintf(stderr, "[%s]\n", str);
 }
 
+/* execute calibration loop */
+static	void	calibrate(struct timex *txp, struct ppsclockev *evp)
+{
+	int		time_state;
+	int		state;
+	state = S_NO_PPS;
+	while ( 1 )
+	{
+		txp->modes = 0;
+		time_state = verbose_adjtimex("monitor: ", 1, txp, evp);
+		print_stats(txp);
+		switch ( state )
+		{
+		case S_NO_PPS:	/* wait for good PPS */
+			update_status(txp, evp,
+				      STA_PPSTIME | STA_PPSFREQ | STA_FLL,
+				      STA_PLL | STA_FREQHOLD);
+			update_offset(txp, evp);
+			if ( (txp->status & STA_PPSSIGNAL) != 0
+			     && (txp->status & PPS_INSANE) == 0
+			     && (txp->offset > -OFFSET_TOLERANCE
+				 && txp->offset < OFFSET_TOLERANCE) )
+			{
+  				state = S_PPS_TIME;
+				status("S_NO_PPS: PPS detected and in range");
+			}
+			break;
+		case S_PPS_TIME:	/* adjust the offset */
+			update_status(txp, evp,
+				      STA_FLL | STA_PPSFREQ,
+				      STA_PPSTIME | STA_FREQHOLD | STA_PLL);
+			update_offset(txp, evp);
+			if ( (evp->tv.tv_usec > 2 * OFFSET_TOLERANCE &&
+			      evp->tv.tv_usec <
+			      1000000 - 2 * OFFSET_TOLERANCE) )
+			{
+				status("S_PPS_TIME: PPS got insane");
+				state = S_NO_PPS;
+			}
+			else if ( (txp->status & STA_PPSSIGNAL) != 0 &&
+				  (txp->status & PPS_INSANE) == 0 &&
+				  txp->offset > -OFFSET_TOLERANCE / 2 &&
+				  txp->offset <= OFFSET_TOLERANCE / 2 &&
+				  txp->stabil <= STABIL_TOLERANCE &&
+				  txp->shift > PPS_SHIFT )
+			{
+  				state = S_PPS_FREQ;
+				status("S_PPS_TIME: PPS offset is low and stable");
+			}
+			break;
+		case S_PPS_FREQ:	/* adjust PLL frequency from PPS */
+			update_status(txp, evp,
+				      STA_FLL | STA_FREQHOLD,
+				      STA_PPSTIME | STA_PPSFREQ | STA_PLL);
+			update_constant(txp, evp);
+			update_offset(txp, evp);
+			if ( txp->shift == PPS_SHIFT ||
+			     txp->offset < -OFFSET_TOLERANCE ||
+			     txp->offset >= OFFSET_TOLERANCE ||
+			     (txp->status & STA_PPSSIGNAL) == 0 ||
+			     (txp->status & ~STA_PPSJITTER & PPS_INSANE) != 0 )
+			{
+				state = S_NO_PPS;
+				status("S_PPS_FREQ: insane PPS signal");
+			}
+			else if ( txp->shift == PPS_SHIFTMAX &&
+				  txp->offset > -OFFSET_TOLERANCE / 4 &&
+				  txp->offset <= OFFSET_TOLERANCE / 4 &&
+				  txp->stabil <= STABIL_TOLERANCE / 2 )
+			{
+				status("PPS_FREQ: best quality PPS signal");
+				state = S_PPS_FLL;
+			}
+			break;
+		case S_PPS_FLL:	/* adjust PLL frequency from PPS */
+			update_status(txp, evp,
+				      STA_FREQHOLD,
+				      STA_PPSTIME | STA_PPSFREQ | STA_FLL |
+				      STA_PLL);
+			update_constant(txp, evp);
+			if ( (txp->status & (STA_PPSSIGNAL)) == 0 ||
+			     (txp->status & ~STA_PPSJITTER & PPS_INSANE) != 0 ||
+			     txp->stabil > STABIL_TOLERANCE ||
+ 			     txp->shift < PPS_SHIFTMAX )
+			{
+				state = S_PPS_FREQ;
+				status("S_PPS_FLL: PPS signal lost quality");
+			}
+			break;
+		}
+		sleep(1);
+	}
+}
+
+/* monitor PPS jitter */
+static	void	monitor_jitter(struct timex *txp, struct ppsclockev *evp)
+{
+	struct timeval	last;
+	long		delta, last_delta;
+
+	last_delta = last.tv_sec = last.tv_usec = 0;
+	txp->modes = 0;
+	while ( 1 )
+	{
+		pps_status("monitor_jitter(): ", 0, 0, txp->status, evp);
+		delta =  evp->tv.tv_usec - last.tv_usec;
+		delta += (evp->tv.tv_sec - last.tv_sec) * 1000000L;
+		last = evp->tv;
+		printf("%6d %8ld\n", evp->serial, delta - last_delta);
+		last_delta = delta;
+		sleep(1);
+	}
+}
+
 /* provide a usage message */
 static	void	usage(void)
 {
 	fprintf(stderr,
 		"Known options are:\n"
+		"\t-C\tdon't adjust time constant in monitoring loop\n"
 		"\t-e\texit - do not enter monitoring loop\n"
+		"\t-j\tmonitor jitter only\n"
 		"\t-k\tkeep settings (when entering monitoring loop)\n"
+		"\t-O\tdon't adjust offset in monitoring loop\n"
 		"\t-p{p|0|n}\tpolarity (of pulse) positive|none|negative\n"
+		"\t-S\tdon't adjust status in monitoring loop\n"
 		);
 }
 
@@ -327,24 +485,34 @@ int	main(int argc, char *argv[])
 	struct serial_struct ss;
 	struct timex 	tx;
 	struct ppsclockev event;
-	int		time_state;
-	int		state;
 	int		ch;
-	int		keep_settings = 0;	/* keep kernel settings */
-	int		monitor = 1;		/* do monitoring loop */
 
 	fprintf(stderr,
 		"WARNING: This program may invalidate your system time!\n");
-	while ( (ch = getopt(argc, argv, "ekp:")) != -1 )
+	printf("(This version has been compiled with"
+#ifndef	STA_NANO
+	       "out"
+#endif
+	       " nanokernel support)\n");
+	while ( (ch = getopt(argc, argv, "CejkOp:S")) != -1 )
 	{
 		switch ( ch )
 		{
-		case 'k':	/* keep settings - don't restart when entering
-				   monitoring loop */
-			keep_settings = 1;
+		case 'C':	/* don't adjust time constant */
+			options = BITS_ON(options, OPT_NO_TIMECONST);
 			break;
 		case 'e':	/* exit - don't enter monitoring loop */
-			monitor = 0;
+			options = BITS_OFF(options, OPT_MONITOR);
+			break;
+		case 'j':	/* monitor jitter only */
+			options = BITS_ON(options, OPT_JITTER_LOOP);
+			break;
+		case 'k':	/* keep settings - don't restart when entering
+				   monitoring loop */
+			options = BITS_OFF(options, OPT_RESET);
+			break;
+		case 'O':	/* don't adjust offset */
+			options = BITS_ON(options, OPT_NO_OFFSET);
 			break;
 		case 'p':	/* select desired edge of the pulse */
 			polarity = optarg[0];
@@ -361,6 +529,9 @@ int	main(int argc, char *argv[])
 				polarity = '0';
 			}
 			break;
+		case 'S':	/* don't adjust time status */
+			options = BITS_ON(options, OPT_NO_STATUS);
+			break;
 		case '?':
 			fprintf (stderr, "Unknown option `-%c'.\n", optopt);
 			usage();
@@ -398,97 +569,14 @@ int	main(int argc, char *argv[])
 	print_header();
 	verbose_adjtimex("init: ", 1, &tx, &event);
 	print_stats(&tx);
-	if ( !keep_settings )
+	if ( options & OPT_RESET )
 	{
 		restart(&tx, &event);
 		print_stats(&tx);
 	}
-	state = S_NO_PPS;
-	while ( monitor )
-	{
-		tx.modes = 0;
-		time_state = verbose_adjtimex("monitor: ", 1, &tx, &event);
-		print_stats(&tx);
-		switch ( state )
-		{
-		case S_NO_PPS:	/* wait for good PPS */
-			update_status(&tx, &event,
-				      STA_PPSTIME | STA_PPSFREQ | STA_FLL,
-				      STA_PLL | STA_FREQHOLD);
-			update_offset(&tx, &event);
-			if ( (tx.status & STA_PPSSIGNAL) != 0
-			     && (tx.status & PPS_INSANE) == 0
-			     && (tx.offset > -OFFSET_TOLERANCE
-				 && tx.offset < OFFSET_TOLERANCE) )
-			{
-  				state = S_PPS_TIME;
-				status("S_NO_PPS: PPS detected and in range");
-			}
-			break;
-		case S_PPS_TIME:	/* adjust the offset */
-			update_status(&tx, &event,
-				      STA_FLL | STA_PPSFREQ,
-				      STA_PPSTIME | STA_FREQHOLD | STA_PLL);
-			update_offset(&tx, &event);
-			if ( (event.tv.tv_usec > 2 * OFFSET_TOLERANCE &&
-			      event.tv.tv_usec <
-			      1000000 - 2 * OFFSET_TOLERANCE) )
-			{
-				status("S_PPS_TIME: PPS got insane");
-				state = S_NO_PPS;
-			}
-			else if ( (tx.status & STA_PPSSIGNAL) != 0 &&
-				  (tx.status & PPS_INSANE) == 0 &&
-				  tx.offset > -OFFSET_TOLERANCE / 2 &&
-				  tx.offset <= OFFSET_TOLERANCE / 2 &&
-				  tx.stabil <= STABIL_TOLERANCE &&
-				  tx.shift > PPS_SHIFT )
-			{
-  				state = S_PPS_FREQ;
-				status("S_PPS_TIME: PPS offset is low and stable");
-			}
-			break;
-		case S_PPS_FREQ:	/* adjust PLL frequency from PPS */
-			update_status(&tx, &event,
-				      STA_FLL | STA_FREQHOLD,
-				      STA_PPSTIME | STA_PPSFREQ | STA_PLL);
-			update_constant(&tx, &event);
-			update_offset(&tx, &event);
-			if ( tx.shift == PPS_SHIFT ||
-			     tx.offset < -OFFSET_TOLERANCE ||
-			     tx.offset >= OFFSET_TOLERANCE ||
-			     (tx.status & STA_PPSSIGNAL) == 0 ||
-			     (tx.status & ~STA_PPSJITTER & PPS_INSANE) != 0 )
-			{
-				state = S_NO_PPS;
-				status("S_PPS_FREQ: insane PPS signal");
-			}
-			else if ( tx.shift == PPS_SHIFTMAX &&
-				  tx.offset > -OFFSET_TOLERANCE / 4 &&
-				  tx.offset <= OFFSET_TOLERANCE / 4 &&
-				  tx.stabil <= STABIL_TOLERANCE / 2 )
-			{
-				status("PPS_FREQ: best quality PPS signal");
-				state = S_PPS_FLL;
-			}
-			break;
-		case S_PPS_FLL:	/* adjust PLL frequency from PPS */
-			update_status(&tx, &event,
-				      STA_FREQHOLD,
-				      STA_PPSTIME | STA_PPSFREQ | STA_FLL |
-				      STA_PLL);
-			update_constant(&tx, &event);
-			if ( (tx.status & (STA_PPSSIGNAL)) == 0 ||
-			     (tx.status & ~STA_PPSJITTER & PPS_INSANE) != 0 ||
-			     tx.stabil > STABIL_TOLERANCE ||
- 			     tx.shift < PPS_SHIFTMAX )
-			{
-				state = S_PPS_FREQ;
-				status("S_PPS_FLL: PPS signal lost quality");
-			}
-			break;
-		}
-		sleep(1);
-	}
+	if ( options & OPT_JITTER_LOOP )
+		monitor_jitter(&tx, &event);
+	if ( options & OPT_MONITOR )
+		calibrate(&tx, &event);
 	return(0);
 }
